import os
import shutil
import stat
import subprocess
import tempfile
import requests
import jenkins
import io
import yaml

from subprocess import CalledProcessError
from charmhelpers.core import hookenv
from charmhelpers.core.host import service_restart
from charmhelpers.core.hookenv import (
    log,
    status_set,
    ERROR,
    WARNING,
    DEBUG,
)
from charms.reactive import (
    hook,
    when,
    when_all,
    when_not,
    set_state,
    get_flags,
    remove_state,
)
from charms.reactive.helpers import data_changed
from charms.layer import scalebot_jenkins

SCALEBOT_URL = "http://localhost:8080"
SCALEBOT_HOME = "/srv/scalebot"
SCALEBOT_REPO = os.path.join(SCALEBOT_HOME, "repo")
PYTHONPATH = os.path.join(SCALEBOT_REPO, "lib")


def _run_cmd_yaml(args):
    """Runs a command, assuming it will send some valid YAML data to stdout,
    and parses its output. All the output will be loaded in memory.
    """

    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    try:
        rstdout, rstderr = proc.communicate(timeout=15)
    except subprocess.TimeoutExpired as exc:
        proc.kill()
        raise exc
    else:
        if proc.returncode != 0:
            raise Exception("Command error: " + str(rstderr))
        return yaml.load(io.BytesIO(rstdout), Loader=yaml.SafeLoader)


@when_not("scalebot.jjb.workaround.installed")
def scalebot_install_jjb_workaround():
    cfgfile = "/etc/default/jenkins"
    tmpfile = "%s.scalebot.tmp" % (cfgfile)

    snippet = [
        "# ScaleBot Workaround for https://storyboard.openstack.org/#!/story/2006489\n",
        'JAVA_ARGS+=" -Dhudson.security.csrf.DefaultCrumbIssuer.EXCLUDE_SESSION_ID=true"\n',
    ]

    with open(cfgfile, "r") as f:
        cfg = f.readlines()
    with open(tmpfile, "w+") as f:
        for line in cfg:
            if line in snippet:
                continue
            f.write("%s" % (line))
        f.writelines(snippet)
    shutil.copymode(cfgfile, tmpfile)
    os.rename(tmpfile, cfgfile)

    # Do a full service restart for Jenkins. We can't do a simple safe
    # restart yet because we need this workaround to connect to the API.
    # We only can do this on deployment, otherwise jobs can be disrupted.
    service_restart("jenkins")

    set_state("scalebot.jjb.workaround.installed")


@when("extension.available")
def configure_jenkins_jobbuilder(extension):
    remove_state("scalebot.configured.jobbuilder")
    status_set("maintenance", "Configuring JobBuilder")
    if not os.path.isdir("/etc/jenkins_jobs"):
        os.mkdir("/etc/jenkins_jobs")

    jj_ini = "/etc/jenkins_jobs/jenkins_jobs.ini"
    if os.path.isfile(jj_ini):
        os.remove(jj_ini)

    config = extension.get_connection_info()
    username = config["admin_username"]
    password = config["admin_password"]

    jenkinsini = open(jj_ini, "w")
    jenkinsini.write(
        """[jenkins]\nuser=%s\n
password=%s
url=%s
"""
        % (username, password, SCALEBOT_URL)
    )
    jenkinsini.close()

    status_set(
        "maintenance", "Configured JobBuilder. Waiting for Scalebot Juju controller."
    )

    log(scalebot_jenkins.if_controller(), DEBUG)

    if scalebot_jenkins.if_controller():
        status_set(
            "maintenance", "Configured JobBuilder and found Scalebot Juju controllers."
        )
        set_state("scalebot.configured.jobbuilder")
        log("scalebot.configured.jobbuilder is set.")

        flags = get_flags()
        log(f"current flags: {flags}", DEBUG)
    else:
        status_set(
            "waiting", "Configured JobBuilder. Waiting for Scalebot Juju controller."
        )


@when("extension.changed")
def unconfigure_jobbuilder():
    remove_state("scalebot.configured.jobbuilder")


@hook("config-changed")
def configure_scalebot_env():
    config = hookenv.config()
    jobset = config.get("scalebot_jobset")
    sbprofile = "/etc/profile.d/scalebot.sh"

    with open(sbprofile, "w") as f:
        f.write("#Export Scalebot required paths\n")
        f.write("export PYTHONPATH=%s\n" % (PYTHONPATH))
        f.write('export SCALEBOT_HOME="%s"\n' % (SCALEBOT_HOME))
        f.write('export SCALEBOT_REPO="%s"\n' % (SCALEBOT_REPO))
        f.write('export SCALEBOT_JOBSET="%s"\n' % (jobset))
        prod = int(config.get("scalebot_production"))
        f.write("export SCALEBOT_PRODUCTION=%d\n" % (prod))

    # We need to restart to get this environment.
    remove_state("scalebot.configured.env")
    set_state("scalebot.written.env")


def scalebot_link_files_into_home():
    # Link files used outside of charm context into $SCALEBOT_HOME
    src = os.path.join(hookenv.charm_dir(), "bin")
    dest = os.path.join(SCALEBOT_HOME, "bin")
    if os.path.islink(dest):
        os.unlink(dest)
    os.symlink(src, dest)


def scalebot_git_config_check():
    required_configs = ["scalebot_config_repo", "scalebot_config_branch"]
    missing_configs = []
    config = hookenv.config()
    for c in required_configs:
        if not config.get(c):
            missing_configs.append(c)
    if len(missing_configs) > 0:
        status_set(
            "blocked",
            "Unable to continue due to missing configs: %s"
            % ", ".join(missing_configs),
        )
        return False
    return True


@when("apt.installed.git")
def scalebot_git_clone():
    if not scalebot_git_config_check():
        return

    config = hookenv.config()
    repo = config.get("scalebot_config_repo")
    branch = config.get("scalebot_config_branch")
    ssh_key = config.get("scalebot_private_ssh_key")

    if not data_changed("scalebot.git.config", [repo, branch, ssh_key]):
        log("Repo config has not changed, nothing to do.")
        return True

    log("Scalebot repo config has changed, (re-)cloning.")
    remove_state("scalebot.configured.repo")
    status_set("maintenance", "Cloning Scalebot repository")

    try:
        shutil.rmtree(SCALEBOT_HOME)
    except FileNotFoundError:
        pass

    os.makedirs(SCALEBOT_HOME)
    shutil.chown(SCALEBOT_HOME, user="jenkins", group="jenkins")
    scalebot_link_files_into_home()

    known_hosts = os.path.join(hookenv.charm_dir(), "files", "known_hosts.lp")

    git_ssh = ""
    keypath = os.path.join(SCALEBOT_HOME, "key")
    if ssh_key:
        log("Writing ssh key to %s" % (keypath))
        with os.fdopen(
            os.open(keypath, os.O_WRONLY | os.O_CREAT, mode=0o600), "w"
        ) as keyfile:
            shutil.chown(keypath, user="jenkins", group="jenkins")
            keyfile.write(ssh_key)
        ssh_cmd = "/usr/bin/ssh -v -i %s -o UserKnownHostsFile=%s" % (
            keypath,
            known_hosts,
        )
        git_ssh = '-c core.sshCommand="%s"' % (ssh_cmd)
    else:
        try:
            os.unlink(keypath)
        except FileNotFoundError:
            pass

    git_cmd = "/usr/bin/git clone %s -b %s %s %s" % (
        git_ssh,
        branch,
        repo,
        os.path.join(SCALEBOT_HOME, "repo"),
    )

    su_git_cmd = ["/bin/su", "-", "jenkins", "-c", git_cmd]
    log("Beginning git clone")
    log(su_git_cmd)
    subprocess.run(su_git_cmd, check=True)

    set_state("scalebot.configured.repo")
    remove_state("scalebot.repo.init")


@when_not("scalebot.repo.init")
def scalebot_repo_init():
    init_path = os.path.join(SCALEBOT_REPO, "scalebot.d", "init")
    if os.path.exists(init_path):
        status_set("maintenance", "Executing %s" % (init_path))
        subprocess.check_call(init_path)
    set_state("scalebot.repo.init")


@hook("config-changed")
def scalebot_git_reconfig():
    config = hookenv.config()
    required_configs = [
        "scalebot_private_ssh_key",
        "scalebot_config_repo",
        "scalebot_config_branch",
    ]
    if any(config.changed(key) for key in required_configs):
        scalebot_git_clone()


@when_all(
    "scalebot.configured.repo",
    "scalebot.configured.jobbuilder",
    "scalebot.configured.juju",
)
def scalebot_ready():
    status_set("active", "ScaleBot is Ready")
    log("Congratulations!! Scalebot is ready. Go testing go!!")


@when_all(
    "scalebot.configured.repo",
    "scalebot.configured.jobbuilder",
    "scalebot.configured.env",
    "scalebot.jjb.workaround.installed",
    "scalebot.jenkins-http-ready",
)
def importjenkins():
    cmd = os.path.join(hookenv.charm_dir(), "actions", "refreshjobs")
    log(cmd, DEBUG)
    try:
        subprocess.run(
            [cmd], check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
        )
    except CalledProcessError as e:
        msg_stdout = e.stdout.decode()
        log(msg_stdout, level=ERROR)


@hook("config-changed")
def scalebot_juju_reconfig():
    config = hookenv.config()
    required_configs = [
        "scalebot_juju_clouds",
        "scalebot_juju_credentials",
        "scalebot_juju_model_defaults",
    ]
    if any(config.changed(key) for key in required_configs):
        scalebot_configure_juju()


@when_all("apt.installed.juju")
def scalebot_configure_juju():
    config = hookenv.config()

    clouds = config.get("scalebot_juju_clouds")
    creds = config.get("scalebot_juju_credentials")
    modeldefaults = config.get("scalebot_juju_model_defaults")
    bootstrapconstraints = config.get("scalebot_juju_bootstrap_constraints")

    if (
        not data_changed("scalebot.juju.config", [clouds, creds])
        and scalebot_jenkins.if_controller()
    ):
        log(
            "Juju controller is available and its config has not changed, nothing to do."
        )

        return True

    if data_changed("scalebot.juju.config", [clouds, creds]):
        log("Juju config has changed, (re-)bootstrapping.")
    elif not scalebot_jenkins.if_controller():
        log("Scalebot Juju controller is not ready, (re-)bootstrapping.")
    else:
        log("Scalebot Juju controller is going to be deployed, (re-)bootstrapping.")

    remove_state("scalebot.configured.juju")
    status_set("maintenance", "Bootstrapping Juju controller")

    # In case we're reconfiguring, destroy any pre-existing
    # controller before re-bootstrapping
    subprocess.run(
        ["/bin/su", "-", "jenkins", "-c", "juju destroy-controller jenkins -y"],
        check=False,
    )

    scalebot_configure_juju_for_lab(clouds, creds, modeldefaults, bootstrapconstraints)


def scalebot_configure_juju_for_lab(clouds, creds, modeldefaults, bootstrapconstraints):
    if clouds:
        cloudsfile = tempfile.NamedTemporaryFile(mode="w", delete=False)
        cloudsfile.write(clouds)
        # Make it readable by user jenkins
        os.chmod(
            cloudsfile.name, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
        )
        cloudsfile.close()
        curr_clouds = _run_cmd_yaml(
            ["/bin/su", "-", "jenkins", "-c", "juju list-clouds --format yaml"]
        )
        if curr_clouds and ("scalebot" in curr_clouds):
            command = "juju update-cloud --client scalebot -f %s" % (cloudsfile.name)
        else:
            command = "juju add-cloud --client scalebot %s" % (cloudsfile.name)
        subprocess.run(["/bin/su", "-", "jenkins", "-c", command], check=True)
        os.unlink(cloudsfile.name)

    if creds:
        credsfile = tempfile.NamedTemporaryFile(mode="w", delete=False)
        credsfile.write(creds)
        # Make it readable by user jenkins
        os.chmod(
            credsfile.name, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
        )
        credsfile.close()
        curr_creds = _run_cmd_yaml(
            ["/bin/su", "-", "jenkins", "-c", "juju show-credentials --format yaml"]
        )
        if (
            curr_creds
            and ("client-credentials" in curr_creds)
            and ("scalebot" in curr_creds["client-credentials"])
        ):
            command = "juju update-credential --client -f %s scalebot" % (
                credsfile.name
            )
        else:
            command = "juju add-credential --client -f %s scalebot" % (credsfile.name)
        subprocess.run(["/bin/su", "-", "jenkins", "-c", command], check=True)
        os.unlink(credsfile.name)

    modeldefaultsparam = ""
    if modeldefaults:
        modeldefaultsfile = tempfile.NamedTemporaryFile(mode="w", delete=False)
        modeldefaultsfile.write(modeldefaults)
        os.chmod(
            modeldefaultsfile.name,
            stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH,
        )
        modeldefaultsfile.close()
        modeldefaultsparam = " --model-default %s" % (modeldefaultsfile.name)

    bconstraints = ""
    if bootstrapconstraints:
        bconstraints = ' --bootstrap-constraints "%s"' % (bootstrapconstraints)

    try:
        subprocess.run(
            [
                "/bin/su",
                "-",
                "jenkins",
                "-c",
                "juju bootstrap --debug scalebot jenkins"
                + modeldefaultsparam
                + bconstraints,
            ],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        log(f"Removing {modeldefaultsfile.name}", DEBUG)
        if modeldefaults:
            os.unlink(modeldefaultsfile.name)
            log(f"Removed {modeldefaultsfile.name}", DEBUG)

        log(
            "Boostraping Scalebot Juju controller looks good. You may wait for it ready."
        )

    except CalledProcessError as e:
        msg_stdout = e.stdout.decode()
        log(msg_stdout, level=ERROR)

        msg_guess = scalebot_jenkins.get_bootstrap_failure_hint(msg_stdout)
        msg = f"Fail to bootstrap a controller. It may be because: {msg_guess}"
        log(msg, level=ERROR)

        # fail to bootstrap a controller is devastating
        # by setting status as "blocked" to help stop the deployment process
        status_set("blocked", msg)

    if scalebot_jenkins.if_controller():
        set_state("scalebot.configured.juju")


@hook("stop")
def scalebot_juju_destroy_controller():
    subprocess.run(
        [
            "/bin/su",
            "-",
            "jenkins",
            "-c",
            "juju destroy-controller --destroy-all-models -y jenkins",
        ]
    )


@when_not("scalebot.jenkins-http-ready")
def jenkins_http_ready():
    rc = 500
    try:
        rc = requests.get(SCALEBOT_URL, timeout=5).status_code
    except (
        requests.HTTPError,
        requests.exceptions.Timeout,
        requests.exceptions.ConnectionError,
    ):
        pass

    # Valid status codes (we can get redirects, login forms, etc.)
    if rc in (200, 300, 301, 302, 403):
        set_state("scalebot.jenkins-http-ready")


# Count the number of jobs currently running on Jenkins and lock it for new
# jobs if zero. This is required to prevent a race condition if new jobs are
# started between running this script and restarting the server.

JK_RESTART_CHECK_SCRIPT = """
import jenkins.model.*
import hudson.model.*

lst = Jenkins.instance.getView('All').getBuilds().findAll() {
    it.getResult().equals(null)
}
if (lst.size() == 0) {
    Jenkins.instance.doQuietDown();
}
println(lst.size());
"""


@when_not("scalebot.configured.env")
@when_all(
    "scalebot.jenkins-http-ready",
    "scalebot.jjb.workaround.installed",
    "scalebot.written.env",
)
@when("extension.available")
def safe_jenkins_service_restart(extension):
    """Restart Jenkins service if no jobs are running.

    Environment variables required by some jobs will not be available until
    we do a full service restart (i.e. finishing Java process and starting
    it again instead of just reloading Jenkins, as done by safe-restart).

    Tests can be lost if we just restart so we need to get the number of
    running jobs (using the Rest API), restart the service and ensure that
    Jenkins is back online before proceeding.
    """

    config = extension.get_connection_info()
    try:
        server = jenkins.Jenkins(
            SCALEBOT_URL,
            username=config["admin_username"],
            password=config["admin_password"],
            timeout=10,
        )
        ret = server.run_script(JK_RESTART_CHECK_SCRIPT)
        if ret and int(ret) == 0:
            log("Restarting Jenkins service")
            service_restart("jenkins")
            remove_state("scalebot.jenkins-http-ready")
            set_state("scalebot.configured.env")
        else:
            log("Jenkins is running some jobs, will restart later")

    except jenkins.JenkinsException as exc:
        log("Jenkins API exception:" + str(exc), level=WARNING)
